/*
 * Copyright (c) 2004-2012 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/* ----------------------------------------------------------------------
 * Functions to init and read the angle from  a KVH5000 gyro
 * ---------------------------------------------------------------------- */


#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

#include <portLib.h>
#include <commonStructLib.h>
#include <taskLib.h>
#include <h2mathLib.h>

#include "serial.h"

#include "gyroPrivate.h"
#include "KVH5000.h"

static void *gyroKVH5000ReadTask(void *);
static BOOL gyroClose(struct GYRO_DATA *);

/* ----------------------------------------------------------------------
 *
 * gyroInit -     Init connection to serial port
 *
 *
 * Returns :            FALSE if failed or TRUE.
 */

struct GYRO_DATA *
gyroInitKVH5000(const char *port, double latitude, double woffset)
{
	struct GYRO_KVH5000_DATA *gyroData;

	if (commonStructCreate(sizeof(struct GYRO_KVH5000_DATA),
		(void **)&gyroData) != OK)
		return FALSE;

	commonStructTake(gyroData);
	if ((gyroData->fd = openSerial(port, 38400)) == -1) {
		warn("openSerial %s", port);
		return FALSE;
	}
	gyroData->gd.wOffset = woffset;
	gyroData->gd.earthSpeed = gyroEarthSpeed(latitude);
	gyroData->gd.close = gyroClose;
	gyroData->stop = 0;
	gyroData->taskId = taskSpawn2(GYRO_KVH5000_READ_TASK_NAME,
	    GYRO_KVH5000_READ_TASK_PRIORITY, VX_FP_TASK,
	    GYRO_KVH5000_READ_TASK_STACK_SIZE, gyroKVH5000ReadTask, gyroData);
	commonStructGive(gyroData);
	if (gyroData->taskId == ERROR) {
		commonStructDelete(gyroData);
		gyroData = NULL;
		return FALSE;
	}
	return (struct GYRO_DATA *)gyroData;
}


/*
 * gyroClose ------------------------------------------------------------
 */
static BOOL
gyroClose(struct GYRO_DATA *p)
{
	struct GYRO_KVH5000_DATA *gyro = (struct GYRO_KVH5000_DATA *)p;

	gyro->stop = 1;
	if (close(gyro->fd) == -1)
		return FALSE;
	commonStructDelete(gyro);
	return TRUE;
}


/*----------------------------------------------------------------------*/

static void
gyroDecode(struct GYRO_KVH5000_DATA *gyro, char *buf, uint32_t date)
{
	double w;
	int val;

	if (strlen(buf) < 20) {
		warnx("gyroDecode: short frame");
		return;
	}
	if (sscanf(buf, "%lf %d", &w, &val) != 2) {
		warnx("gyroDecode: invalid frame");
		return;
	}

	w = DEG_TO_RAD(w) - gyro->gd.earthSpeed;

	commonStructTake(gyro);
	if (!val) {
		gyro->dataValid = 0;
		goto done;
	}
	gyro->gd.date = date;
	gyro->dataValid = 1;
	gyroIntegrate((struct GYRO_DATA *)gyro, w);
done:
	commonStructGive(gyro);
}


static void *
gyroKVH5000ReadTask(void *param)
{
	static char buf[32];
	struct GYRO_KVH5000_DATA *gyro = (struct GYRO_KVH5000_DATA *)param;
	ssize_t nread;
	int rptr, rem;
	char *p;

	rptr = 0;
	while (!gyro->stop) {
		/* read some characters */
		nread = read(gyro->fd, buf + rptr, sizeof(buf) - rptr);
		if (nread < 0)
			err(2, "read");
		/* look for CR */
		p = strchr(buf+rptr, '\r');
		if (p != NULL) {
			/* found eom */
			*p = '\0';
			gyroDecode(gyro, buf, getTime());
			/* Handle remaining characters */
			rem = nread - (p - (buf + rptr)) - 1;
			/* recopy them at begining of buf */
			memmove(buf, p+1, rem);
			/* and clear the rest */
			memset(buf+rem, '0', sizeof(buf) - rem);
			rptr = rem;
		} else {
			rptr +=  nread;
			if (rptr == sizeof(buf)) {
				warnx("can't find EOM");
				rptr = 0;
			}
		}
	}
	return NULL;
}
